"""

    Greynir: Natural language processing for Icelandic

    Settings module

    Copyright © 2025 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


    This module reads and interprets the GreynirCorrect.conf
    configuration file. The file can include other files using the $include
    directive, making it easier to arrange configuration sections into logical
    and manageable pieces.

    Sections are identified like so: [ section_name ]

    Comments start with # signs.

    Sections are interpreted by section handlers.

"""

from typing import Any, Dict, Iterator, List, Mapping, Optional, Set, Tuple

import os
import threading
from collections import defaultdict
from importlib.resources import files

from reynir.bindb import GreynirBin
from reynir.bintokenizer import StateDict


ErrorFormTuple = Tuple[str, str, int, str, str]
# (lemma, id, cat, correct_word_form, tag, eink, malsnid, stafs, aslatt, beyg)
RitmyndirTuple = Tuple[str, int, str, str, str, int, str, str, str, str]
# (ritregla/spelling rule, category, other detail ) }
DetailsTuple = Tuple[str, str, str]
# A set of all strings that should be interpreted as True
TRUE = frozenset(("true", "True", "1", "yes", "Yes"))
# Einkunn value from Ritmyndir mapped to error code
R_EINKUNN: Mapping[int, str] = {
    0: "R000",
    1: "R001",  # Not an error
    2: "R002",
    3: "R003",
    4: "R004",
    5: "R005",
}


class ConfigError(Exception):
    """Exception class for configuration errors"""

    def __init__(self, s: str) -> None:
        super().__init__(s)
        self.fname: Optional[str] = None
        self.line = 0

    def set_pos(self, fname: str, line: int) -> None:
        """Set file name and line information, if not already set"""
        if not self.fname:
            self.fname = fname
            self.line = line

    def __str__(self) -> str:
        """Return a string representation of this exception"""
        s = Exception.__str__(self)
        if not self.fname:
            return s
        return "File {0}, line {1}: {2}".format(self.fname, self.line, s)


class LineReader:
    """Read lines from a text file, recognizing $include directives"""

    def __init__(
        self,
        fname: str,
        *,
        package_name: Optional[str] = None,
        outer_fname: Optional[str] = None,
        outer_line: int = 0
    ) -> None:
        self._fname = fname
        self._package_name = package_name
        self._line = 0
        self._inner_rdr: Optional[LineReader] = None
        self._outer_fname = outer_fname
        self._outer_line = outer_line

    def fname(self) -> str:
        """The name of the file being read"""
        return self._fname if self._inner_rdr is None else self._inner_rdr.fname()

    def line(self) -> int:
        """The number of the current line within the file"""
        return self._line if self._inner_rdr is None else self._inner_rdr.line()

    def lines(self) -> Iterator[str]:
        """Generator yielding lines from a text file"""
        self._line = 0
        try:
            if self._package_name:
                ref = files(self._package_name).joinpath(self._fname)
                stream = ref.open("rb")
            else:
                stream = open(self._fname, "rb")
            with stream as inp:
                # Read config file line-by-line from the package resources
                accumulator = ""
                for b in inp:
                    # We get byte strings; convert from utf-8 to Python strings
                    s = b.decode("utf-8")
                    self._line += 1
                    if s.rstrip().endswith("\\"):
                        # Backslash at end of line: continuation in next line
                        accumulator += s.strip()[:-1]
                        continue
                    if accumulator:
                        # Add accumulated text from preceding
                        # backslash-terminated lines, but drop leading whitespace
                        s = accumulator + s.lstrip()
                        accumulator = ""
                    # Check for include directive: $include filename.txt
                    if s.startswith("$") and s.lower().startswith("$include "):
                        iname = s.split(maxsplit=1)[1].strip()
                        # Do some path magic to allow the included path
                        # to be relative to the current file path, or a
                        # fresh (absolute) path by itself
                        head, _ = os.path.split(self._fname)
                        iname = os.path.join(head, iname)
                        rdr = self._inner_rdr = LineReader(
                            iname,
                            package_name=self._package_name,
                            outer_fname=self._fname,
                            outer_line=self._line,
                        )
                        yield from rdr.lines()
                        self._inner_rdr = None
                    else:
                        yield s
                if accumulator:
                    # Catch corner case where last line of file ends with a backslash
                    yield accumulator
        except (IOError, OSError):
            if self._outer_fname:
                # This is an include file within an outer config file
                c = ConfigError(
                    "Error while opening or reading include file '{0}'".format(
                        self._fname
                    )
                )
                c.set_pos(self._outer_fname, self._outer_line)
            else:
                # This is an outermost config file
                c = ConfigError(
                    "Error while opening or reading config file '{0}'".format(
                        self._fname
                    )
                )
            raise c


class AllowedMultiples:
    def __init__(self) -> None:
        self.SET: Set[str] = set()

    def add(self, word: str) -> None:
        self.SET.add(word)


class WrongCompounds:
    def __init__(self) -> None:
        # Dictionary structure: dict { wrong_compound : "right phrase" }
        self.DICT: Dict[str, Tuple[str, ...]] = {}

    def add(self, word: str, parts: Tuple[str, ...]) -> None:
        if word in self.DICT:
            raise ConfigError("Multiple definition of '{0}' in wrong_compounds section".format(word))
        assert isinstance(parts, tuple)
        self.DICT[word] = parts


class SplitCompounds:
    # Dict of the form { first_part : set(second_part_stem) }
    def __init__(self) -> None:
        self.DICT: Dict[str, Set[str]] = defaultdict(set)

    def add(self, first_part: str, second_part_stem: str) -> None:
        if first_part in self.DICT and second_part_stem in self.DICT[first_part]:
            raise ConfigError(
                "Multiple definition of '{0}' in split_compounds section".format(first_part + " " + second_part_stem)
            )
        self.DICT[first_part].add(second_part_stem)


class UniqueErrors:
    # Dictionary structure: dict { wrong_word : (tuple of right words) }
    def __init__(self) -> None:
        self.DICT: Dict[str, Tuple[str, ...]] = dict()

    def add(self, word: str, corr: Tuple[str, ...]) -> None:
        if word in self.DICT:
            raise ConfigError("Multiple definition of '{0}' in unique_errors section".format(word))
        self.DICT[word] = corr


class MultiwordErrors:
    def __init__(self) -> None:
        # Dictionary structure: dict { phrase tuple: error specification }
        # List of tuples of multiword error phrases and their word category lists
        self.LIST: List[Tuple[Tuple[str, ...], str, List[str]]] = []
        # Parsing dictionary keyed by first word of phrase
        self.DICT: StateDict = defaultdict(list)
        # Error dictionary, { phrase : (error_code, right_phrase, right_parts_of_speech) }
        self.ERROR_DICT: Dict[Tuple[str, ...], str] = dict()

    def add(self, words: Tuple[str, ...], error: str) -> None:
        if words in self.ERROR_DICT:
            raise ConfigError("Multiple definition of '{0}' in multiword_errors section".format(" ".join(words)))
        self.ERROR_DICT[words] = error

        # Add to phrase list
        ix = len(self.LIST)

        a = error.split(",")
        if len(a) != 2:
            raise ConfigError("Expected two comma-separated parameters within $error()")
        code = a[0].strip()
        replacement = a[1].strip().split()

        # Append the phrase and the error specification in tuple form
        self.LIST.append((words, code, replacement))

        # Dictionary structure: dict { firstword: [ (restword_list, phrase_index) ] }
        self.DICT[words[0]].append((list(words[1:]), ix))

    def get_phrase(self, ix: int) -> Tuple[str, ...]:
        """Return the original phrase with index ix"""
        return self.LIST[ix][0]

    def get_phrase_length(self, ix: int) -> int:
        """Return the count of words in the original phrase with index ix"""
        return len(self.LIST[ix][0])

    def get_code(self, ix: int) -> str:
        """Return the error code with index ix"""
        return self.LIST[ix][1]

    def get_replacement(self, ix: int) -> List[str]:
        """Return the replacement phrase with index ix"""
        return self.LIST[ix][2]


class TabooWords:
    def __init__(self) -> None:
        # Dictionary structure: dict { taboo_word : (suggested_replacement, explanation) }
        self.DICT: Dict[str, Tuple[str, str]] = {}

    def add(self, word: str, replacement: str, explanation: str) -> None:
        if word in self.DICT:
            raise ConfigError("Multiple definition of '{0}' in taboo_words section".format(word))
        db = GreynirBin.get_db()
        a = word.split("_")
        _, m = db.lookup_g(a[0])
        if not m or (len(a) >= 2 and all(mm.ordfl != a[1] for mm in m)):
            raise ConfigError("The taboo word '{0}' is not found in BÍN".format(word))
        self.DICT[word] = (replacement, explanation)


class ToneOfVoiceWords:
    def __init__(self) -> None:
        # Dictionary structure: dict { tone_of_voice : (suggested_replacement, explanation) }
        self.DICT: Dict[str, Tuple[str, str]] = {}

    def add(self, word: str, replacement: str, explanation: str) -> None:
        if word in self.DICT:
            raise ConfigError("Multiple definition of '{0}' in tone_of_voice section".format(word))
        db = GreynirBin.get_db()
        a = word.split("_")
        _, m = db.lookup_g(a[0])
        if not m or (len(a) >= 2 and all(mm.ordfl != a[1] for mm in m)):
            raise ConfigError("The word '{0}' is not found in BÍN".format(word))
        self.DICT[word] = (replacement, explanation)


class ToneOfVoicePatterns:
    def __init__(self) -> None:
        # A path to a python module containing third party tone of voice patterns
        self.PATH: str = ""

    def add(self, fpath: str) -> None:
        self.PATH = fpath  # type: ignore[reportConstantRedefinition]


class Suggestions:
    def __init__(self) -> None:
        # Dictionary structure: dict { bad_word : [ suggested_replacements ] }
        self.DICT: Dict[str, List[str]] = {}

    def add(self, word: str, replacements: List[str]) -> None:
        if word in self.DICT:
            raise ConfigError("Multiple definition of '{0}' in suggestions section".format(word))
        self.DICT[word] = replacements


class CapitalizationErrors:
    def __init__(self) -> None:
        # Set of wrongly capitalized words
        self.SET: Set[str] = set()
        # Reverse capitalization (íslendingur -> Íslendingur, Danskur -> danskur)
        self.SET_REV: Set[str] = set()

    def emulate_case(self, s: str, template: str) -> str:
        """Return the string s but emulating the case of the template
        (lower/upper/capitalized)"""
        if template.isupper():
            return s.upper()
        if template and template[0].isupper():
            return s.capitalize()
        return s

    def reverse_capitalization(self, word: str, *, split_on_hyphen: bool = False) -> str:
        """Return a word with its capitalization reversed (lower <-> upper case)"""
        if split_on_hyphen and "-" in word:
            # 'norður-kórea' -> 'Norður-Kórea'
            return "-".join(self.reverse_capitalization(part) for part in word.split("-"))
        if word.islower():
            # Lowercase word
            word_rev = word.capitalize()
        elif word.isupper() and len(word) > 1:
            # Multi-letter uppercase acronym
            word_rev = word.capitalize()
        elif word[0].isupper() and word[1:].islower():
            # Uppercase word
            word_rev = word.lower()
        else:
            raise ConfigError("'{0}' cannot have mixed capitalization".format(word))
        return word_rev

    def add(self, word: str) -> None:
        """Add the given (wrongly capitalized) word stem to the stem set"""
        # We support compound words such as 'félags- og barnamálaráðherra' here
        split_on_hyphen = False
        if " " in word:
            prefix, suffix = word.rsplit(" ", maxsplit=1)
            prefix += " "
        else:
            prefix, suffix = "", word
            # Split_on_hyphen is True for e.g. 'norður-kórea' and 'nýja-sjáland'
            split_on_hyphen = "-" in word
        db = GreynirBin().get_db()
        # The suffix may not be in BÍN except as a compound, and in that
        # case we want its hyphenated lemma
        suffix_rev = self.reverse_capitalization(suffix, split_on_hyphen=split_on_hyphen)
        _, m = db.lookup_g(suffix_rev)
        # Only consider lemmas
        STOFN_NOT_IN_ORDMYNDIR = ("Bandaríki",)
        m = [mm for mm in m if (mm.stofn == mm.ordmynd) or (mm.stofn in STOFN_NOT_IN_ORDMYNDIR)]
        if not m:
            raise ConfigError(
                "No BÍN meaning for '{0}' (from error word '{1}') in capitalization_errors section".format(
                    suffix_rev, word
                )
            )
        if not prefix:
            # This might be something like 'barnamálaráðherra' which comes out
            # with a lemma of 'barnamála-ráðherra'
            word = self.emulate_case(m[0].stofn, template=word)
        else:
            # This might be something like 'félags- og barnamálaráðherra' which comes out
            # with a lemma of 'félags- og barnamála-ráðherra'
            word = prefix + m[0].stofn
        if word in self.SET:
            raise ConfigError("Multiple definition of '{0}' in capitalization_errors section".format(word))
        # Construct the reverse casing of the word
        word_rev = self.reverse_capitalization(word, split_on_hyphen=split_on_hyphen)
        # Add the word and its reverse case to the set of errors
        self.SET.add(word)
        self.SET_REV.add(word_rev)


class OwForms:
    def __init__(self) -> None:
        # dict { wrong_word_form : (lemma, correct_word_form, id, cat, tag) }
        self.DICT: Dict[str, Tuple[str, str, int, str, str]] = dict()

    def contains(self, word: str) -> bool:
        """Check whether the word form is in the error forms dictionary,
        either in its original casing or in a lower case form"""
        d = self.DICT
        if word.islower():
            return word in d
        return word in d or word.lower() in d

    def add(self, wrong_form: str, meaning: Tuple[str, str, int, str, str]) -> None:
        self.DICT[wrong_form] = meaning

    def get_lemma(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][0]

    def get_correct_form(self, wrong_form: str) -> str:
        """Return a corrected form of the given word, attempting
        to emulate the lower/upper/title case of the word"""
        # First, try the original casing of the wrong form
        c = self.DICT.get(wrong_form)
        if c is not None:
            # Found it: we're done
            return c[1]
        # Lookup a lower case version
        c = self.DICT.get(wrong_form.lower())
        if c is None:
            # Not found: can't correct
            return wrong_form
        form = c[1]
        if wrong_form.istitle():
            return form.title()
        if wrong_form.isupper():
            return form.upper()
        return form

    def get_id(self, wrong_form: str) -> int:
        return self.DICT[wrong_form][2]

    def get_category(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][3]

    def get_tag(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][4]


class CIDErrorForms:
    def __init__(self) -> None:
        # dict { wrong_word_form : (lemma, correct_word_form, id, cat, tag) }
        self.DICT: Dict[str, ErrorFormTuple] = dict()

    def contains(self, word: str) -> bool:
        """Check whether the word form is in the error forms dictionary,
        either in its original casing or in a lower case form"""
        d = self.DICT
        return word in d or word.lower() in d

    def add(self, wrong_form: str, meaning: ErrorFormTuple) -> None:
        self.DICT[wrong_form] = meaning

    def get_lemma(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][0]

    def get_correct_form(self, wrong_form: str) -> str:
        """Return a corrected form of the given word, attempting
        to emulate the lower/upper/title case of the word"""
        # First, try the original casing of the wrong form
        c = self.DICT.get(wrong_form)
        if c is not None:
            # Found it: we're done
            return c[1]
        # Lookup a lower case version
        c = self.DICT.get(wrong_form.lower())
        if c is None:
            # Not found: can't correct
            return wrong_form
        form = c[1]
        if wrong_form.istitle():
            return form.title()
        if wrong_form.isupper():
            return form.upper()
        return form

    def get_id(self, wrong_form: str) -> int:
        return self.DICT[wrong_form][2]

    def get_category(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][3]

    def get_tag(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][4]


class CDErrorForms:
    def __init__(self) -> None:
        # dict { wrong_word_form : (lemma, correct_word_form, id, cat, tag) }
        self.DICT: Dict[str, ErrorFormTuple] = dict()

    def contains(self, word: str) -> bool:
        """Check whether the word form is in the error forms dictionary,
        either in its original casing or in a lower case form"""
        d = self.DICT
        if word.islower():
            return word in d
        return word in d or word.lower() in d

    def add(self, wrong_form: str, meaning: ErrorFormTuple) -> None:
        self.DICT[wrong_form] = meaning

    def get_lemma(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][0]

    def get_correct_form(self, wrong_form: str) -> str:
        """Return a corrected form of the given word, attempting
        to emulate the lower/upper/title case of the word"""
        # First, try the original casing of the wrong form
        c = self.DICT.get(wrong_form)
        if c is not None:
            # Found it: we're done
            return c[1]
        # Lookup a lower case version
        c = self.DICT.get(wrong_form.lower())
        if c is None:
            # Not found: can't correct
            return wrong_form
        form = c[1]
        if wrong_form.istitle():
            return form.title()
        if wrong_form.isupper():
            return form.upper()
        return form

    def get_id(self, wrong_form: str) -> int:
        return self.DICT[wrong_form][2]

    def get_category(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][3]

    def get_tag(self, wrong_form: str) -> str:
        return self.DICT[wrong_form][4]


class Morphemes:
    def __init__(self) -> None:
        # dict { morpheme : [ preferred PoS ] }
        self.BOUND_DICT: Dict[str, List[str]] = {}
        # dict { morpheme : [ excluded PoS ] }
        self.FREE_DICT: Dict[str, List[str]] = {}

    def add(self, morph: str, boundlist: List[str], freelist: List[str]) -> None:
        if not boundlist:
            raise ConfigError("A definition of allowed PoS is necessary with morphemes")
        self.BOUND_DICT[morph] = boundlist
        # The freelist may be empty
        self.FREE_DICT[morph] = freelist


class Ritmyndir:
    # dict { wrong_word_form : (lemma, id, cat, correct_word_form, tag, eink, malsnid, stafs, aslatt, beyg) }
    # þurrð;10963;kvk;þurðar;þurrðar;EFET;0;URE;;;;1745-1745;KLIM
    # þurrka;425063;so;þurkaði;þurrkaði;;4;VILLA;R4RR;;;;SKOLAVERK
    def __init__(self) -> None:
        self.DICT: Dict[str, RitmyndirTuple] = dict()

    def contains(self, word: str) -> bool:
        """Check whether the word form is in the Ritmyndir dictionary"""
        return word in self.DICT or word.lower() in self.DICT

    def get_lemma(self, wrong_form: str) -> Optional[str]:
        return self.get_entry(wrong_form, 0)

    def get_id(self, wrong_form: str) -> Optional[int]:
        return self.get_entry(wrong_form, 1)

    def get_cat(self, wrong_form: str) -> str:
        return self.get_entry(wrong_form, 2)

    def get_correct_form(self, wrong_form: str) -> str:
        return self.get_entry(wrong_form, 3)

    def get_tag(self, wrong_form: str) -> str:
        return self.get_entry(wrong_form, 4)

    def get_eink(self, wrong_form: str) -> int:
        eink = self.get_entry(wrong_form, 3)
        if not eink:
            eink = 1
        return eink

    def get_malsnid(self, wrong_form: str) -> str:
        return self.get_entry(wrong_form, 6).split(",")[0]

    def get_stafs(self, wrong_form: str) -> str:
        return self.get_entry(wrong_form, 7).split(",")[0]

    def get_aslatt(self, wrong_form: str) -> str:
        return self.get_entry(wrong_form, 8).split(",")[0]

    def get_beyg(self, wrong_form: str) -> str:
        return self.get_entry(wrong_form, 9).split(",")[0]

    def get_code(self, wrong_form: str) -> str:
        code = self.get_stafs(wrong_form)
        if not code:
            code = self.get_aslatt(wrong_form)
        if not code:
            code = self.get_beyg(wrong_form)
        if not code:
            code = self.get_malsnid(wrong_form)
        if not code:
            code = R_EINKUNN[self.get_eink(wrong_form)]
        if not code:
            code = "R001"
        return code

    def get_entry(self, wrong_form: str, index: int) -> Any:
        entry = self.DICT.get(wrong_form, self.DICT.get(wrong_form.lower()))
        return entry[index] if entry else None

    def add(self, wrong_form: str, details: RitmyndirTuple) -> None:
        # TODO Same ritmynd can occur multiple times in the data from different references, how to handle?
        # TODO Also check if the same ritmynd has many different corrections in the data,
        # so we don't just overwrite former values. The DICT entries can be a list of corrections, using defaultdict()
        self.DICT[wrong_form] = details


class RitmyndirDetails:
    def __init__(self) -> None:
        # "Ritmyndir error code" : ("ritregla/spelling rule", "category", "other detail" ) }
        self.DICT: Dict[str, DetailsTuple] = dict()

    def add(self, code: str, details: DetailsTuple) -> None:
        self.DICT[code] = details


class IecNonwords:
    def __init__(self) -> None:
        # Dictionary structure: dict { wrong_word : (tuple of right words) }
        self.DICT: Dict[str, Tuple[str, ...]] = dict()

    def add(self, word: str, corr: Tuple[str, ...]) -> None:
        if word in self.DICT:
            # Happens in the data, just skip it
            # raise ConfigError(
            #    "Multiple definition of '{0}' in IEC nonwords data".format(word)
            # )
            return
        self.DICT[word] = corr


class Icesquer:
    def __init__(self) -> None:
        # Dictionary structure: dict { wrong_word : (tuple of right words) }
        self.DICT: Dict[str, Tuple[str, ...]] = dict()

    def add(self, word: str, corr: Tuple[str, ...]) -> None:
        if word in self.DICT:
            raise ConfigError(f"Multiple definition of '{word}' in IEC nonwords data")
        self.DICT[word] = corr


class WrongFormers:
    def __init__(self) -> None:
        # Dictionary structure: dict { wrong_word : right_word }
        self.DICT: Dict[str, str] = dict()

    def add(self, word: str, corr: str) -> None:
        if word in self.DICT:
            raise ConfigError(f"Multiple definition of '{word}' in WrongFormers")
        self.DICT[word] = corr


class WrongFormersCID:
    def __init__(self) -> None:
        # Dictionary structure: dict { wrong_word : right_word }
        self.DICT: Dict[str, str] = dict()

    def add(self, word: str, corr: str) -> None:
        if word in self.DICT:
            raise ConfigError(f"Multiple definition of '{word}' in WrongFormersCID")
        self.DICT[word] = corr


class Settings:
    """Global settings"""

    def __init__(self):
        self.allowed_multiples = AllowedMultiples()
        self.wrong_compounds = WrongCompounds()
        self.split_compounds = SplitCompounds()
        self.unique_errors = UniqueErrors()
        self.multiword_errors = MultiwordErrors()
        self.taboo_words = TabooWords()
        self.suggestions = Suggestions()
        self.capitalization_errors = CapitalizationErrors()
        self.ow_forms = OwForms()
        self.cid_error_forms = CIDErrorForms()
        self.cd_error_forms = CDErrorForms()
        self.morphemes = Morphemes()
        self.ritmyndir = Ritmyndir()
        self.ritmyndir_details = RitmyndirDetails()
        self.iec_nonwords = IecNonwords()
        self.icesquer = Icesquer()
        self.tone_of_voice_words = ToneOfVoiceWords()
        self.tone_of_voice_patterns = ToneOfVoicePatterns()
        self.wrong_formers = WrongFormers()
        self.wrong_formers_cid = WrongFormersCID()

    _lock = threading.Lock()
    loaded = False
    DEBUG = os.environ.get("DEBUG", "").strip() in TRUE

    # Configuration settings from the GreynirCorrect.conf file

    def _handle_settings(self, s: str) -> None:
        """Handle config parameters in the settings section"""
        a: List[str] = s.lower().split("=", maxsplit=1)
        par = a[0].strip().lower()
        val = a[1].strip()
        try:
            if par == "debug":
                Settings.DEBUG = val in TRUE  # type: ignore[reportConstantRedefinition]
            else:
                raise ConfigError("Unknown configuration parameter '{0}'".format(par))
        except ValueError:
            raise ConfigError("Invalid parameter value: {0} = {1}".format(par, val))

    def _handle_allowed_multiples(self, s: str) -> None:
        """Handle config parameters in the allowed_multiples section"""
        assert s
        if len(s.split()) != 1:
            raise ConfigError("Only one word per line allowed in allowed_multiples section")
        if s in self.allowed_multiples.SET:
            raise ConfigError("'{0}' is repeated in allowed_multiples section".format(s))
        self.allowed_multiples.add(s)

    def _handle_wrong_compounds(self, s: str) -> None:
        """Handle config parameters in the wrong_compounds section"""
        a = s.lower().split(",", maxsplit=1)
        if len(a) != 2:
            raise ConfigError("Expected comma between compound word and its parts")
        word = a[0].strip().strip('"')
        parts = a[1].strip().strip('"').split()
        if not word:
            raise ConfigError("Expected word before the comma in wrong_compounds section")
        if len(parts) < 2:
            raise ConfigError("Missing word part(s) in wrong_compounds section")
        if len(word.split()) != 1:
            raise ConfigError("Multiple words not allowed before comma in wrong_compounds section")
        self.wrong_compounds.add(word, tuple(parts))

    def _handle_split_compounds(self, s: str) -> None:
        """Handle config parameters in the split_compounds section"""
        parts = s.split()
        if len(parts) != 2:
            raise ConfigError("Missing word part(s) in split_compounds section")
        self.split_compounds.add(parts[0], parts[1])

    def _handle_unique_errors(self, s: str) -> None:
        """Handle config parameters in the unique_errors section"""
        a = s.lower().split(",", maxsplit=1)
        if len(a) != 2:
            raise ConfigError("Expected comma between error word and its correction")
        word = a[0].strip()
        if len(word) < 3:
            raise ConfigError("Expected nonempty word before comma in unique_errors section")
        if word[0] != '"' or word[-1] != '"':
            raise ConfigError("Expected word in double quotes in unique_errors section")
        word = word[1:-1]
        corr = a[1].strip()
        if len(corr) < 3:
            raise ConfigError("Expected nonempty word after comma in unique_errors section")
        if corr[0] != '"' or corr[-1] != '"':
            raise ConfigError("Expected word in double quotes after comma in unique_errors section")
        corr = corr[1:-1]
        corr_t = tuple(corr.split())
        if not word:
            raise ConfigError("Expected word before the comma in unique_errors section")
        if len(word.split()) != 1:
            raise ConfigError("Multiple words not allowed before the comma in unique_errors section")
        self.unique_errors.add(word, corr_t)

    def _handle_capitalization_errors(self, s: str) -> None:
        """Handle config parameters in the capitalization_errors section"""
        self.capitalization_errors.add(s)

    def _handle_taboo_words(self, s: str) -> None:
        """Handle config parameters in the taboo_words section"""
        # Start by parsing explanation string off the end (right hand side), if present
        lquote = s.find('"')
        rquote = s.rfind('"')
        if (lquote >= 0) != (rquote >= 0):
            raise ConfigError("Explanation string for taboo word should be enclosed in double quotes")
        if lquote >= 0:
            # Obtain explanation from within quotes
            explanation = s[lquote + 1 : rquote].strip()
            s = s[:lquote].rstrip()
        else:
            # No explanation
            explanation = ""
        if not s:
            raise ConfigError("Expected taboo word and a suggested replacement")
        a = s.lower().split()
        if len(a) > 2:
            raise ConfigError("Expected taboo word and a suggested replacement")
        taboo = a[0].strip()
        if len(a) == 2:
            replacement = a[1].strip()
        else:
            replacement = taboo
        # Check all replacement words, which are separated by slashes '/'
        if any(r.count("_") != 1 for r in replacement.split("/")):
            raise ConfigError("Suggested replacement(s) should include a word category (_xx)")
        self.taboo_words.add(taboo, replacement, explanation)

    def _handle_tone_of_voice_words(self, s: str) -> None:
        """Handle config parameters in the tone_of_voice section."""
        # Start by parsing explanation string off the end (right hand side), if present
        lquote = s.find('"')
        rquote = s.rfind('"')
        if (lquote >= 0) != (rquote >= 0):
            raise ConfigError("Explanation string for a word should be enclosed in double quotes")
        if lquote >= 0:
            # Obtain explanation from within quotes
            explanation = s[lquote + 1 : rquote].strip()
            s = s[:lquote].rstrip()
        else:
            # No explanation
            explanation = ""
        if not s:
            raise ConfigError("Expected a word to flag and a suggested replacement")
        a = s.split()  # not lower() here
        if len(a) > 2:
            raise ConfigError("Expected a word to flag and a suggested replacement")
        flagged_word = a[0].strip()
        if len(a) == 2:
            replacement = a[1].strip()
        else:
            replacement = flagged_word
        # Check all replacement words, which are separated by slashes '/'
        if any(r.count("_") != 1 for r in replacement.split("/")):
            raise ConfigError("Suggested replacement(s) should include a word category (_xx)")
        self.tone_of_voice_words.add(flagged_word, replacement, explanation)

    def _handle_tone_of_voice_patterns(self, s: str) -> None:
        """Handle module path for external patterns."""
        # the string includes quotes, let's remove them
        p = s.split("=")[1].strip().strip('"')
        # Should only be a path to a python file
        if not os.path.exists(p):
            raise ConfigError("Not a valid path. Expected a path to a Python file")
        self.tone_of_voice_patterns.add(p)

    def _handle_suggestions(self, s: str) -> None:
        """Handle config parameters in the suggestions section"""
        a = s.lower().split()
        if len(a) < 2:
            raise ConfigError("Expected flagged word and at least one suggested replacement")
        if any(w.count("_") != 1 for w in a[1:]):
            raise ConfigError("Suggested replacements should include word category (_xx)")
        self.suggestions.add(a[0].strip(), [w.strip() for w in a[1:]])

    def _handle_multiword_errors(self, s: str) -> None:
        """Handle config parameters in the multiword_errors section"""
        a = s.lower().split("$error", maxsplit=1)
        if len(a) != 2:
            raise ConfigError("Expected phrase followed by $error(...)")
        phrase = tuple(a[0].strip().split())
        if len(phrase) < 2:
            raise ConfigError("Multiword phrase must contain at least two words")
        error = a[1].strip()
        if len(error) < 3:
            raise ConfigError("Incomplete error specification for multiword phrase")
        if error[0] != "(" or error[-1] != ")":
            raise ConfigError("Error specification should be enclosed in parentheses")
        self.multiword_errors.add(phrase, error[1:-1])

    def _handle_ow_forms(self, s: str) -> None:
        """Handle config parameters in the ow_forms section"""
        split = s.strip().split(";")
        if len(split) != 6:
            raise ConfigError("Expected wrong form;lemma;correct form;id;category;tag")
        wrong_form = split[0].strip()
        correct_form = split[2].strip()
        if wrong_form == correct_form:
            return
            # !!! TODO: Should do this:
            # raise ConfigError(
            #     "Wrong form identical to correct form for '{0}'".format(wrong_form)
            # )
        meaning: ErrorFormTuple = (
            split[1].strip(),  # Lemma (stofn)
            correct_form,  # Correct form (ordmynd)
            int(split[3]),  # Id (utg)
            split[4].strip(),  # Category (ordfl)
            split[5].strip(),  # Tag (beyging)
        )
        self.ow_forms.add(wrong_form, meaning)

    def _handle_error_forms(self, s: str) -> None:
        """Handle config parameters in the error_forms section"""
        split = s.strip().split(";")
        if len(split) != 7:
            raise ConfigError("Expected wrong form;lemma;correct form;id;category;tag;errortype")
        wrong_form = split[0].strip()
        correct_form = split[2].strip()
        if wrong_form == correct_form:
            raise ConfigError("Wrong form identical to correct form for '{0}'".format(wrong_form))
        meaning: ErrorFormTuple = (
            split[1].strip(),  # Lemma (stofn)
            correct_form,  # Correct form (ordmynd)
            int(split[3]),  # Id (utg)
            split[4].strip(),  # Category (ordfl)
            split[5].strip(),  # Tag (beyging)
        )
        etype = split[6].strip()
        if etype == "cid":
            self.cid_error_forms.add(wrong_form, meaning)  # context-independent errors
        elif etype == "cd":
            self.cd_error_forms.add(wrong_form, meaning)  # context-dependent errors
        else:
            raise ConfigError("Wrong error type given, expected 'cid' or 'cd'")

    def _handle_morphemes(self, s: str) -> None:
        """Process the contents of the [morphemes] section"""
        freelist: List[str] = []
        boundlist: List[str] = []
        spl = s.split()
        if len(spl) < 2:
            raise ConfigError("Expected at least a prefix and an attachment specification")
        m = spl[0]
        for pos in spl[1:]:
            if pos:
                if pos.startswith("+"):
                    boundlist.append(pos[1:])
                elif pos.startswith("-"):
                    freelist.append(pos[1:])
                else:
                    raise ConfigError("Attachment specification should start with '+' or '-'")
        self.morphemes.add(m, boundlist, freelist)

    def _handle_ritmyndir(self, s: str) -> None:
        """Handle data from Ritmyndir in Stórasnið in BÍN/DIM"""
        split = s.strip().split(";")
        if len(split) != 13:
            raise ConfigError(
                "Expected lemma, id, cat, wrong_word_form, correct_word_form, tag, eink, malsnid, stafs, aslatt, beyg, age, ref"
            )
        ref = split[12].strip()
        if "SAGA" in ref or ref in {"MILTON", "HALLGP-4", "KLIM", "ONP"}:
            # Skipping errors from very old references, don't represent errors in Modern Icelandic
            return
        wrong_form = split[3].strip()
        correct_form = split[4].strip()
        if wrong_form == correct_form:
            return
        if wrong_form.lower() == correct_form.lower():
            # TODO Skipping capitalization errors for now
            return
        # (lemma, id, cat, correct_word_form, tag, eink, malsnid, stafs, aslatt, beyg)
        meaning: RitmyndirTuple = (
            split[0].strip(),  # Lemma
            int(split[1].strip()),  # id
            split[2].strip(),  # cat
            correct_form,  # correct_word_form
            split[5].strip(),  # tag
            int(split[6].strip()),  # eink
            split[7].strip(),  # malsnid
            split[8].strip(),  # stafs
            split[9].strip(),  # aslatt
            split[10].strip(),  # beyg
        )
        self.ritmyndir.add(wrong_form, meaning)

    def _handle_ritmyndir_details(self, s: str) -> None:
        """Handle data on Ritmyndir categories, including references to the Icelandic Standards"""
        # "Ritmyndir error code" : ("ritregla/spelling rule", "category", "other detail" )
        split = s.split(":", maxsplit=1)
        code = split[0].strip().strip('"')
        dsplit = split[1].split('", "')
        rule = dsplit[0].strip().strip('"')
        cat = dsplit[1].strip().strip('"')
        det = dsplit[2].strip().strip('"')
        self.ritmyndir_details.DICT[code] = (rule, cat, det)

    def _handle_iec_nonwords(self, s: str) -> None:
        """Handle config parameters in the Icelandic Error Corpus Nonwords"""
        a = s.lower().split("\t")
        if len(a) != 2:
            # Happens in the data, just skip it
            # raise ConfigError("Expected tab between error word and its correction")
            return
        word = a[0].strip()
        if len(word) < 1:
            raise ConfigError("Expected nonempty word before comma in unique_errors section")
        corr = a[1].strip()
        if len(corr) < 1:
            raise ConfigError("Expected nonempty word after comma in unique_errors section")
        corr_t = tuple(corr.split())
        if not word:
            raise ConfigError("Expected word before the comma in unique_errors section")
        if len(word.split()) != 1:
            # Happens in the data, just skip it
            return
            # raise ConfigError(
            #    "Multiple words not allowed before the comma in unique_errors section"
            # )
        self.iec_nonwords.add(word, corr_t)

    def _handle_icesquer(self, s: str) -> None:
        """Handle config parameters in the Icelandic Error Corpus Nonwords"""
        a = s.lower().split("\t")
        if len(a) != 2:
            # Happens in the data, just skip it
            # raise ConfigError("Expected tab between error word and its correction")
            return
        word = a[0].strip()
        if len(word) < 1:
            raise ConfigError("Expected nonempty word before comma in unique_errors section")
        corr = a[1].split(";")[0].strip()  # TODO Only the first value for now
        if len(corr) < 1:
            raise ConfigError("Expected nonempty word after comma in unique_errors section")
        corr_t = tuple(corr.split())
        if not word:
            raise ConfigError("Expected word before the comma in unique_errors section")
        if len(word.split()) != 1:
            # Happens in the data, just skip it
            return
            # raise ConfigError(
            #    "Multiple words not allowed before the comma in unique_errors section"
            # )
        self.icesquer.add(word, corr_t)

    def _handle_wrong_formers(self, s: str) -> None:
        """Handle config parameters in the wrong_formers section"""
        a = s.lower().split(",", maxsplit=1)
        if len(a) != 2:
            raise ConfigError("Expected comma between the word and its correction")
        word = a[0].strip().strip('"')
        correction = a[1].strip().strip('"')
        if not word:
            raise ConfigError("Expected word before the comma in wrong_formers section")
        if not correction:
            raise ConfigError("Expected word after the comma in wrong_formers section")
        if len(word.split()) != 1 or len(correction.split()) != 1:
            raise ConfigError("Expected one word on each side in wrong_formers")
        self.wrong_formers.add(word, correction)

    def _handle_wrong_formers_cid(self, s: str) -> None:
        """Handle config parameters in the wrong_formers section"""
        a = s.lower().split(",", maxsplit=1)
        if len(a) != 2:
            raise ConfigError("Expected comma between the word and its correction")
        word = a[0].strip().strip('"')
        correction = a[1].strip().strip('"')
        if not word:
            raise ConfigError("Expected word before the comma in wrong_formers section")
        if not correction:
            raise ConfigError("Expected word after the comma in wrong_formers section")
        if len(word.split()) != 1 or len(correction.split()) != 1:
            raise ConfigError("Expected one word on each side in wrong_formers")
        self.wrong_formers_cid.add(word, correction)

    def read(self, fname: str, external: bool = False) -> None:
        """Read configuration file"""

        with Settings._lock:
            CONFIG_HANDLERS = {
                "settings": Settings._handle_settings,
                "allowed_multiples": Settings._handle_allowed_multiples,
                "wrong_compounds": Settings._handle_wrong_compounds,
                "split_compounds": Settings._handle_split_compounds,
                "unique_errors": Settings._handle_unique_errors,
                "capitalization_errors": Settings._handle_capitalization_errors,
                "taboo_words": Settings._handle_taboo_words,
                "tone_of_voice_words": Settings._handle_tone_of_voice_words,
                "tone_of_voice_patterns": Settings._handle_tone_of_voice_patterns,
                "suggestions": Settings._handle_suggestions,
                "multiword_errors": Settings._handle_multiword_errors,
                "morphemes": Settings._handle_morphemes,
                "ow_forms": Settings._handle_ow_forms,
                "error_forms": Settings._handle_error_forms,
                "auto_ow": Settings._handle_ow_forms,
                "auto_error": Settings._handle_error_forms,
                "iec_nonwords": Settings._handle_iec_nonwords,
                "icesquer": Settings._handle_icesquer,
                "ritmyndir": Settings._handle_ritmyndir,
                "ritmyndir_details": Settings._handle_ritmyndir_details,
                "wrong_formers": Settings._handle_wrong_formers,
                "wrong_formers_ci": Settings._handle_wrong_formers_cid,
            }
            handler = None  # Current section handler

            rdr = None

            try:
                # If an external path is given, use it to read the file
                package_name = None if external else __name__.split(".")[0]
                rdr = LineReader(fname, package_name=package_name)
                for s in rdr.lines():
                    # Ignore comments
                    ix = s.find("#")
                    if ix >= 0:
                        s = s[0:ix]
                    s = s.strip()
                    if not s:
                        # Blank line: ignore
                        continue
                    if s[0] == "[" and s[-1] == "]":
                        # New section
                        section = s[1:-1].strip().lower()
                        if section in CONFIG_HANDLERS:
                            handler = CONFIG_HANDLERS[section]
                            continue
                        raise ConfigError("Unknown section name '{0}'".format(section))
                    if handler is None:
                        raise ConfigError("No handler for config line '{0}'".format(s))
                    # Call the correct handler depending on the section
                    try:
                        handler(self, s)
                    except ConfigError as e:
                        # Add file name and line number information to the exception
                        # if it's not already there
                        e.set_pos(rdr.fname(), rdr.line())
                        raise e

            except ConfigError as e:
                # Add file name and line number information to the exception
                # if it's not already there
                if rdr:
                    e.set_pos(rdr.fname(), rdr.line())
                raise e

            Settings.loaded = True



"""

    Greynir: Natural language processing for Icelandic

    Wrapper functions module

    Copyright © 2025 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


    This module exposes functions to return corrected strings given an input text.
    The following options are defined:

    input:  Defines the input. Can be a string or an iterable of strings, such
            as a file object.
    format: Defines the output format. String.
            text: Output is returned as a corrected version of the input.
            json: Output is returned as a JSON string.
            csv:  Output is returned in a csv format.
            m2:   Output is returned in the M2 format, see https://github.com/nusnlp/m2scorer
                  The output is as follows:
                  S <tokenized system output for sentence 1>
                  A <token start offset> <token end offset>|||<error type>|||<correction1>||<correction2||..||correctionN|||<required>|||<comment>|||<annotator id>
    all_errors: Defines the level of correction. If False, only token-level annotation is carried out.
                If True, sentence-level annotation is carried out.
    annotate_unparsed_sentences: If True, sentences that cannot be parsed are annotated as errors as a whole.
    annotations: If True, can all error annotations are returned at the end of the output. Works with format text.
    generate_suggestion_list: If True, the annotation can in certain cases contain a list of possible corrections, for the user to pick from.
    suppress_suggestions: If True, more farfetched automatically retrieved corrections are rejected and no error is added.
    ignore_wordlist: The value is a set of strings, a whitelist. Each string is a word that should not be marked as an error or corrected.
    one_sent: Defines input as containing only one sentence.
    ignore_rules: A list of error codes that should be ignored in the annotation process.
"""

from __future__ import annotations

from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union, cast
from typing_extensions import TypedDict

import argparse
import json
import logging
from dataclasses import dataclass
from functools import partial

from tokenizer import TOK, calculate_indexes, detokenize, normalized_text_from_tokens, text_from_tokens
from tokenizer.definitions import AmountTuple, NumberTuple

from .errtokenizer import CorrectionPipeline, CorrectToken, Error, load_config
from .readability import FleschKincaidFeedback, FleschKincaidScorer, RareWordsFinder
from .annotation import Annotation
from .checker import AnnotatedSentence, CheckResult, GreynirCorrect
from .classifier import SentenceClassifier

log = logging.getLogger(__name__)


class AnnTokenDict(TypedDict, total=False):
    """Type of the token dictionaries returned from check_errors()"""

    # Token kind
    k: int
    # Token text
    x: str
    # Original text of token
    o: str
    # Character offset of token, indexed from the start of the checked text
    i: int


class AnnDict(TypedDict):
    """A single annotation, as returned by the Yfirlestur.is API"""

    start: int
    end: int
    start_char: int
    end_char: int
    code: str
    text: str
    detail: Optional[str]
    suggest: Optional[str]


class AnnResultDict(TypedDict):
    """The annotation result for a sentence"""

    original: str
    corrected: str
    annotations: List[AnnDict]
    tokens: List[AnnTokenDict]


# File types for UTF-8 encoded text files
ReadFile = argparse.FileType("r", encoding="utf-8")
WriteFile = argparse.FileType("w", encoding="utf-8")

# Configure our JSON dump function
json_dumps = partial(json.dumps, ensure_ascii=False, separators=(",", ":"))


def quote(s: str) -> str:
    """Return the string s within double quotes, and with any contained
    backslashes and double quotes escaped with a backslash"""
    if not s:
        return '""'
    return '"' + s.replace("\\", "\\\\").replace('"', '\\"') + '"'


def val(t: CorrectToken, quote_word: bool = False) -> Union[None, str, float, Tuple[Any, ...], Sequence[Any]]:
    """Return the value part of the token t"""
    if t.val is None:
        return None
    if t.kind in {TOK.WORD, TOK.PERSON, TOK.ENTITY}:
        # No need to return list of meanings
        return None
    if t.kind in {TOK.PERCENT, TOK.NUMBER, TOK.CURRENCY}:
        return cast(NumberTuple, t.val)[0]
    if t.kind == TOK.AMOUNT:
        num, iso, _, _ = cast(AmountTuple, t.val)
        if quote_word:
            # Format as "1234.56|USD"
            return '"{0}|{1}"'.format(num, iso)
        return num, iso
    if t.kind == TOK.S_BEGIN:
        return None
    if t.kind == TOK.PUNCTUATION:
        punct = t.punctuation
        return quote(punct) if quote_word else punct
    if quote_word and t.kind in {
        TOK.DATE,
        TOK.TIME,
        TOK.DATEABS,
        TOK.DATEREL,
        TOK.TIMESTAMP,
        TOK.TIMESTAMPABS,
        TOK.TIMESTAMPREL,
        TOK.TELNO,
        TOK.NUMWLETTER,
        TOK.MEASUREMENT,
    }:
        # Return a |-delimited list of numbers
        return quote("|".join(str(v) for v in cast(Iterable[Any], t.val)))
    if quote_word and isinstance(t.val, str):
        return quote(t.val)
    return t.val


@dataclass
class CorrectedSentence:
    tokens: List[CorrectToken]
    parsed: bool
    annotations: Optional[List[Annotation]] = None

    @staticmethod
    def from_parser_sentence(sentence: AnnotatedSentence) -> CorrectedSentence:
        """Create a CorrectedSentence from a parser sentence"""
        tokens = sentence.correct_tokens
        parsed = False
        if sentence.tree is not None and sentence.terminals is not None:
            # Successfully parsed: use the text from the terminals (where available)
            # since we have more info there, for instance on em/en dashes.
            # Create a map of token indices to corresponding terminal text
            parsed = True
            token_map = {t.index: t.text for t in sentence.terminals}
            for ix, tok in enumerate(tokens):
                tok.txt = token_map.get(ix, tok.txt)
        return CorrectedSentence(tokens=tokens, parsed=parsed, annotations=sentence.annotations)

    def filter_annotations(self, ignore_rules: frozenset[str]) -> None:
        """Remove ignored annotations"""
        # If there are no annotations, or no ignore rules, we return early
        if self.annotations is None or not ignore_rules:
            return
        ann = [a for a in self.annotations if a.code not in ignore_rules]
        self.annotations = ann

    def original_str(self) -> str:
        """Return the original text of the sentence"""
        return "".join(t.original or t.txt for t in self.tokens)

    def corrected_str(self, apply_annotations: bool = False) -> str:
        """Return the corrected text of the sentence, with annotation suggestions applied if requested"""
        if apply_annotations:
            return fully_correct_sentence(self.tokens, self.annotations or [])
        return detokenize(self.tokens)


@dataclass
class ParseResultStats:
    num_sentences: int
    num_parsed: int
    num_tokens: int
    ambiguity: float
    parse_time: float


@dataclass
class CorrectionResult:
    sentences: List[CorrectedSentence]
    parse_result_stats: Optional[ParseResultStats] = None
    flesch_result: Optional[Tuple[float, FleschKincaidFeedback]] = None
    rare_words: Optional[List[Tuple[str, float]]] = None

    def filter_annotations(self, ignore_rules: frozenset[str]) -> None:
        """Remove ignored annotations"""
        for sent in self.sentences:
            sent.filter_annotations(ignore_rules)


class GreynirCorrectAPI:
    """A high level api for correcting Icelandic texts"""

    def __init__(
        self,
        gc: GreynirCorrect,
        sentence_prefilter: Optional[SentenceClassifier] = None,
        do_flesch: bool = False,
        rare_word_analyzer: Optional[RareWordsFinder] = None,
        do_grammar_check: bool = True,
    ):
        self.gc = gc
        self.do_grammar_check = do_grammar_check
        self.sentence_prefilter = sentence_prefilter
        self.do_flesch = do_flesch
        # If it's defined, it will be used
        self.rare_word_analyzer = rare_word_analyzer

    @staticmethod
    def from_options(**options: Any) -> GreynirCorrectAPI:
        """Create a GreynirCorrectAPI from the given options"""
        settings = load_config(options.pop("tov_config", None))
        do_flesch_analysis = bool(options.pop("flesch", False))
        do_rare_word_analysis = bool(options.pop("rare_words", False))
        pipeline = CorrectionPipeline(
            "",
            settings,
            **options,
        )
        gc = GreynirCorrect(
            settings=settings,
            pipeline=pipeline,
            **options,
        )
        rare_word_analyzer = RareWordsFinder() if do_rare_word_analysis else None
        do_grammar_check = options.get("all_errors", True)
        sentence_prefilter: SentenceClassifier | None = None
        if options.get("sentence_prefilter", False):
            # Only construct the classifier model if we need it
            from .classifier import SentenceClassifier
            sentence_prefilter = SentenceClassifier()
        return GreynirCorrectAPI(
            gc,
            sentence_prefilter=sentence_prefilter,
            do_flesch=do_flesch_analysis,
            rare_word_analyzer=rare_word_analyzer,
            do_grammar_check=do_grammar_check,
        )

    def _correct_spelling(
        self, text: Iterable[str], ignore_rules: Optional[frozenset[str]] = None, suppress_suggestions: bool = False
    ) -> Iterable[CorrectToken]:
        """Correct the token-level errors in the text"""
        # TODO: The pipeline needs a refactoring.
        # We use some hacks here to avoid having to rewrite the pipeline at this point.
        self.gc.pipeline._text_or_gen = text  # type: ignore[reportPrivateUsage]
        self.gc.pipeline._ignore_rules = ignore_rules or frozenset()  # type: ignore[reportPrivateUsage]
        self.gc.pipeline._suppress_suggestions = suppress_suggestions  # type: ignore[reportPrivateUsage]
        return self.gc.pipeline.tokenize()  # type: ignore

    def _sentence_contains_error(self, corrected_tokens: Iterable[CorrectToken]) -> bool:
        """Classify a sentence as probably correct or not."""
        if self.sentence_prefilter is None:
            raise ValueError("Sentence classifier not initialized - did you forget to set sentence_prefilter=True?")
        original_sentence = "".join(t.original or t.txt for t in corrected_tokens)
        return self.sentence_prefilter.classify(original_sentence)

    def _correct_grammar(self, corrected_tokens: Iterable[CorrectToken]) -> CheckResult:
        results = self.gc.parse_all_tokens(corrected_tokens)
        return results

    def correct(
        self, text: Iterable[str], ignore_rules: Optional[frozenset[str]] = None, suppress_suggestions: bool = False
    ) -> CorrectionResult:
        """Correct the input text by first correcting spelling and then grammatical errors."""
        corrected_tokens = self._correct_spelling(
            text, ignore_rules=ignore_rules, suppress_suggestions=suppress_suggestions
        )
        # Convert the tokens to a list, so it can be reused - this must be done at some point anyway
        corrected_tokens = list(corrected_tokens)
        flesch_result = None
        if self.do_flesch:
            flesch_score = FleschKincaidScorer.get_score_from_stream(corrected_tokens)
            flesch_feedback = FleschKincaidScorer.get_feedback(flesch_score)
            flesch_result = (flesch_score, flesch_feedback)
        rare_words = None
        if self.rare_word_analyzer is not None:
            # TODO: support setting max_words and low_prob_cutoff on function call
            rare_words = self.rare_word_analyzer.get_rare_words_from_stream(
                [tok for tok in corrected_tokens if tok.kind == TOK.WORD], max_words=10, low_prob_cutoff=0.00000005
            )
        if not self.do_grammar_check:
            # Only run the spelling correction
            return CorrectionResult(
                sentences=[CorrectedSentence(tokens=corrected_tokens, parsed=False)],
                flesch_result=flesch_result,
                rare_words=rare_words,
            )
        # Only run the sentence classifier if we should
        if self.sentence_prefilter is not None:
            # Check if the sentence contains an error
            # TODO: We will want to chunk the input into sentences and run the classifier on each sentence
            if not self._sentence_contains_error(corrected_tokens):
                # The sentence is probably correct, so we skip the rest of the processing
                return CorrectionResult(
                    sentences=[CorrectedSentence(tokens=corrected_tokens, parsed=False)],
                    flesch_result=flesch_result,
                    rare_words=rare_words,
                )
            # The sentence is probably incorrect, so we continue with the full grammar check
        # Run the full grammar check
        check_result = self._correct_grammar(corrected_tokens)
        corrected_sentences = [CorrectedSentence.from_parser_sentence(sentence=s) for s in check_result["sentences"]]
        result = CorrectionResult(
            sentences=corrected_sentences,
            flesch_result=flesch_result,
            rare_words=rare_words,
            parse_result_stats=ParseResultStats(
                num_sentences=check_result["num_sentences"],
                num_parsed=check_result["num_parsed"],
                num_tokens=check_result["num_tokens"],
                ambiguity=check_result["ambiguity"],
                parse_time=check_result["parse_time"],
            ),
        )
        # Filter annotations based on ignore rules
        result.filter_annotations(ignore_rules=ignore_rules or frozenset())
        return result


def check_errors(**options: Any) -> str:
    """Return a string in the chosen format and correction level
    using the spelling and grammar checker"""
    text: str | List[str] = options.pop("input", "")
    if not text:
        return ""
    all_errors: bool = options.pop("all_errors", True)
    format: str = options.pop("format", "json")
    spaced: bool = options.pop("spaced", False)
    normalize: bool = options.pop("normalize", False)
    annotations: bool = options.pop("annotations", False)
    print_all: bool = options.pop("print_all", False)
    ignore_rules: frozenset[str] = options.pop("ignore_rules", frozenset())
    suppress_suggestions: bool = options.pop("suppress_suggestions", False)
    api = GreynirCorrectAPI.from_options(**options)
    if isinstance(text, str):
        text = [text]
    results = api.correct(text, ignore_rules=ignore_rules, suppress_suggestions=suppress_suggestions)
    text_results = ""
    if all_errors:
        text_results = format_output(results=results, format=format, print_annotations=annotations)
    else:
        text_results = format_spelling(
            results=results,
            format=format,
            spaced=spaced,
            normalize=normalize,
            print_annotations=annotations,
            print_all=print_all,
        )
    # Add the Flesch score and rare words at the end of the output
    if results.flesch_result is not None:
        flesch_score, flesch_feedback = results.flesch_result
        text_results += f"\nFlesch-Kincaid score: {flesch_score:.2f} ({flesch_feedback})"
    if results.rare_words is not None:
        text_results += "\nRare words:\n"
        for word, _prob in results.rare_words:
            text_results += f"\t{word}\n"
    return text_results


def format_spelling(
    results: CorrectionResult,
    format: str = "json",
    spaced: bool = False,
    normalize: bool = False,
    print_annotations: bool = False,
    print_all: bool = False,
) -> str:
    # Initialize sentence accumulator list
    # Function to convert a token list to output text
    if spaced:
        if normalize:
            to_text = normalized_text_from_tokens
        else:
            to_text = text_from_tokens
    else:
        to_text = partial(detokenize, normalize=True)
    unisum: List[str] = []
    allsum: List[str] = []
    annlist: List[str] = []
    for sent in results.sentences:
        sent_tokens = sent.tokens
        if format == "text":
            txt = to_text(sent_tokens)
            if print_annotations:
                for t in sent_tokens:
                    if t.error:
                        annlist.append(str(t.error))
                if annlist and not print_all:
                    txt = txt + "\n" + "\n".join(annlist)
                    annlist = []
            unisum.append(txt)
            continue
        for t in sent_tokens:
            if format == "csv":
                if t.txt:
                    allsum.append(
                        "{0},{1},{2},{3}".format(
                            t.kind,
                            quote(t.txt),
                            val(t, quote_word=True) or '""',
                            quote(str(t.error) if t.error else ""),
                        )
                    )
                elif t.kind == TOK.S_END:
                    # Indicate end of sentence
                    allsum.append('0,"",""')
            elif format == "json":
                # Output the tokens in JSON format, one line per token
                d: Dict[str, Any] = dict(k=TOK.descr[t.kind])
                if t.txt:
                    d["t"] = t.txt
                v = val(t)
                if t.kind not in {TOK.WORD, TOK.PERSON, TOK.ENTITY} and v is not None:
                    d["v"] = v
                if isinstance(t.error, Error):
                    d["e"] = t.error.to_dict()
                allsum.append(json_dumps(d))
        if allsum:
            unisum.extend(allsum)
            allsum = []
    if print_all:
        # We want the annotations at the bottom
        unistr = " ".join(unisum)
        if annlist:
            unistr = unistr + "\n" + "\n".join(annlist)
    else:
        unistr = "\n".join(unisum)
    return unistr


def fully_correct_sentence(tokens: List[CorrectToken], annotations: List[Annotation]) -> str:
    """Fully correct a sentence by applying the annotations to the tokens and return the text"""
    # Sort in ascending order by token start index, and then by end index
    # (more narrow/specific annotations before broader ones)
    annotations.sort(key=lambda ann: (ann.start, ann.end), reverse=True)

    # Generate a sentence with all corrections applied
    full_correction_toks = tokens[:]
    for ann in annotations:
        if ann.suggest is None:
            # Nothing to correct with, nothing we can do
            continue
        full_correction_toks[ann.start].txt = ann.suggest
        if ann.end > ann.start:
            # Annotation spans many tokens
            # "Okkur börnunum langar í fisk"
            # "Leita að kílómeter af féinu" → leita að kílómetri af fénu → leita að kílómetra af fénu
            # "dást af þeim" → "dást að þeim"
            # Single-token annotations for this span have already been handled
            # Only case is one ann, many toks in toklist
            # Give the first token the correct value
            # Delete the other tokens
            del full_correction_toks[ann.start + 1 : ann.end + 1]
    fully_corrected_sentence = detokenize(full_correction_toks)
    return fully_corrected_sentence


def format_output(
    results: CorrectionResult,
    format: str,
    print_annotations: bool = False,
) -> str:
    """
    Format grammar analysis results in the given format.

    `sentence_results` is a list of individual sentences and their analysis
    `format_type` is the output format to use, one of 'text', 'json', 'csv', 'm2'
    `extra_text_options` takes extra options for the text format. Ignored for other formats.
    """

    if format == "text":
        return format_text(results, print_annotations=print_annotations)
    elif format == "json":
        return format_json(results)
    elif format == "csv":
        return format_csv(results)
    elif format == "m2":
        return format_m2(results)

    raise ValueError(f"Tried to format with invalid format: {format}")


def format_text(results: CorrectionResult, print_annotations: bool = False) -> str:
    output: List[str] = []
    for result in results.sentences:
        txt = fully_correct_sentence(result.tokens, result.annotations or [])

        if print_annotations:
            txt = txt + "\n" + "\n".join(str(ann) for ann in result.annotations or [])
        output.append(txt)

    return "\n".join(output)


def format_json(results: CorrectionResult) -> str:
    formatted_sentences: List[str] = []

    offset = 0
    for result in results.sentences:
        # Calculate the character offsets of the tokens in the original text
        # This returns [0, ... , len(x)]
        char_indexes, _ = calculate_indexes(result.tokens, last_is_end=True)

        # Convert the annotations to a standard format before encoding in JSON
        formatted_annotations: List[AnnDict] = [
            AnnDict(
                # Start token index of this annotation
                start=ann.start,
                # End token index (inclusive)
                end=ann.end,
                # Character offset of the start of the annotation in the original text
                start_char=char_indexes[ann.start] + offset,
                # Character offset of the end of the annotation in the original text
                # (inclusive, i.e. the offset of the last character)
                end_char=char_indexes[ann.end + 1] - 1 + offset,
                code=ann.code,
                text=ann.text,
                detail=ann.detail or "",
                suggest=ann.suggest or "",
            )
            for ann in result.annotations or []
        ]
        # The offset for the next sentence is the end of the last annotation + 1
        ard = AnnResultDict(
            original=result.original_str(),
            corrected=result.corrected_str(apply_annotations=True),
            tokens=[
                AnnTokenDict(k=tok.kind, x=tok.txt, o=tok.original or "", i=index + offset)
                for tok, index in zip(result.tokens, char_indexes)
            ],
            annotations=formatted_annotations,
        )

        formatted_sentences.append(json.dumps(ard, ensure_ascii=False))
        # The offset for the next sentence needs to be increased by the length of this (original) sentence
        offset += char_indexes[-1]

    return "\n".join(formatted_sentences)


def format_csv(results: CorrectionResult) -> str:
    accumul: List[str] = []
    for result in results.sentences:
        for ann in result.annotations or []:
            accumul.append(
                "{},{},{},{},{},{}".format(
                    ann.code,
                    ann.original,
                    ann.suggest,
                    ann.start,
                    ann.end,
                    ann.suggestlist,
                )
            )
    return "\n".join(accumul)


def format_m2(results: CorrectionResult) -> str:
    accumul: List[str] = []
    for result in results.sentences:
        accumul.append("S {0}".format(" ".join(t.txt for t in result.tokens)))
        for ann in result.annotations or []:
            accumul.append(
                "A {0} {1}|||{2}|||{3}|||REQUIRED|||-NONE-|||0".format(ann.start, ann.end, ann.code, ann.suggest)
            )
        accumul.append("")
    return "\n".join(accumul)



"""

    Greynir: Natural language processing for Icelandic

    Sentence classifier module

    Copyright © 2025 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


    This module contains a sentence classifier that classifies sentences
    based on whether they probably contain a grammatical error or not.
    The classifier uses a neural network for the classification and does not
    require a full sentence parse by Greynir.

    The goal is that the classifier runs much faster than the full parse and
    can therefore be used as a filter to skip parsing sentences that are
    probably correct.

    Note: The classifier has a nonzero error rate. Depending on your use case
    this may be acceptable or not.

"""

from typing import Any, List, Union, overload


class SentenceClassifier:
    _model_name = "mideind/yfirlestur-icelandic-classification-byt5"
    _true_label = "1"
    _domain_prefix = "has_error "

    def __init__(self) -> None:
        try:
            from transformers import pipeline  # type: ignore
        except ImportError:
            import warnings

            warningtext = (
                "Tried to import the classifier module without the required packages installed.\n"
                "The required packages are in the 'sentence_classifier' extra\n"
                "Run 'pip install reynir-correct[sentence_classifier] to install them\n"
                "\n"
                "Alternatively, install the packages directly with\n"
                "'pip install datasets transformers torch' or similar.\n"
            )
            warnings.warn(warningtext)
            raise ImportError(warningtext)
        self.pipe: Any = pipeline(
            "text2text-generation",
            model=self._model_name,
            tokenizer="google/byt5-base",
        )

    @overload
    def classify(self, text: str) -> bool:
        ...

    @overload
    def classify(self, text: List[str]) -> List[bool]:
        ...

    def classify(self, text: Union[str, List[str]]) -> Union[List[bool], bool]:
        """Classify a sentence or sentences.
        For each sentence, return true if the sentence probably contains an error."""
        if isinstance(text, str):
            text = [text]

        pipe_result = self.pipe([self._domain_prefix + t for t in text])
        result: List[bool] = [r["generated_text"] == self._true_label for r in pipe_result]

        return result[0] if len(result) == 1 else result


def _main() -> None:
    classifier = SentenceClassifier()

    many_sents = [
        "Þesi settníng er ekki rét sfsett.",
        "Þessi setning er rétt stafsett.",
    ]
    many_results = classifier.classify(many_sents)
    one_sent = "Þesi er vavasöm."
    one_result = classifier.classify(one_sent)

    print(f"Sentence: {many_sents[0]}")
    print(f"Result:   {many_results[0]}")

    print(f"Sentence: {many_sents[1]}")
    print(f"Result:   {many_results[1]}")

    print(f"Sentence: {one_sent}")
    print(f"Result:   {one_result}")


if __name__ == "__main__":
    _main()



#!/usr/bin/env python
"""

    Greynir: Natural language processing for Icelandic

    Spelling and grammar checking module

    Copyright © 2025 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


    This is an executable program wrapper (main module) for the GreynirCorrect
    package. It can be used to invoke the corrector from the command line,
    or via fork() or exec(), with the command 'correct'. The main() function
    of this module is registered as a console_script entry point in setup.py.

"""

from typing import Dict, Union

import argparse
import sys

from .wrappers import check_errors

# File types for UTF-8 encoded text files
ReadFile = argparse.FileType("r", encoding="utf-8")
WriteFile = argparse.FileType("w", encoding="utf-8")

# Define the command line arguments

parser = argparse.ArgumentParser(description="Corrects Icelandic text")

parser.add_argument(
    "infile",
    nargs="?",
    type=ReadFile,
    default=sys.stdin,
    help="UTF-8 text file to correct",
)
parser.add_argument(
    "outfile",
    nargs="?",
    type=WriteFile,
    default=sys.stdout,
    help="UTF-8 output text file",
)

parser.add_argument(
    "--suppress_suggestions",
    "-ss",
    action="store_true",
    help="Suppress more agressive error suggestions",
)

parser.add_argument("--spaced", "-sp", help="Separate tokens with spaces", action="store_true")

# Determines the output format
parser.add_argument(
    "--format",
    "-f",
    nargs="?",
    type=str,
    default="text",
    help="""Determine output format.
text: Corrected text only.
csv: One token per line in CSV format.
json: One token per line in JSON format.
m2: M2 format, GEC standard.""",
)

# Determines whether we supply only token-level annotations or also sentence-level annotations
parser.add_argument(
    "--all_errors",
    "-a",
    help="Annotate both grammar and spelling errors",
    action="store_true",
)

# Add --grammar for compatibility; works the same as --all_errors
parser.add_argument(
    "--grammar",
    "-g",
    help="Annotate both grammar and spelling errors",
    action="store_true",
)

# Add --json for compatibility; works the same as --format=json
parser.add_argument(
    "--json",
    "-j",
    help="Output in JSON format",
    action="store_true",
)

# Add --csv for compatibility; works the same as --format=csv
parser.add_argument(
    "--csv",
    "-c",
    help="Output in CSV format",
    action="store_true",
)

# Add --normalize
parser.add_argument(
    "--normalize",
    "-n",
    help="Normalize punctuation",
    action="store_true",
)

parser.add_argument(
    "--sentence_prefilter",
    help="""Run a heuristic filter on sentences to determine whether they are probably
correct. Probably correct sentences will not go through the full parsing process.""",
    action="store_true",
)
parser.add_argument(
    "--flesch",
    help="Calculate Flesch readability score for the input text",
    action="store_true",
)
parser.add_argument(
    "--rare_words",
    help="Show rare words in the input text",
    action="store_true",
)

parser.add_argument(
    "--tov_config",
    nargs=1,
    type=str,
    help="""Add additional use-specific rules in a configuration file to check for custom
tone-of-voice issues. Uses the same format as the default GreynirCorrect.conf file""",
    default=None,
)

parser.add_argument(
    "--suggest_not_correct",
    help="Instead of directly changing the text, some stylistic errors are presented as suggestions only.",
    action="store_true",
    default=False,
)


def from_args(args: argparse.Namespace) -> Dict[str, Union[str, bool]]:
    """Fill options with information from args"""
    format = args.format
    if args.json or args.grammar:  # The --grammar option implies --json
        format = "json"
    elif args.csv:
        format = "csv"
    return {
        "input": args.infile,
        "suppress_suggestions": args.suppress_suggestions,
        "format": format,
        "spaced": args.spaced,
        "normalize": args.normalize,
        "all_errors": args.all_errors or args.grammar,
        "sentence_prefilter": args.sentence_prefilter,
        "tov_config": args.tov_config,
        "suggest_not_correct": args.suggest_not_correct,
        "flesch": args.flesch,
        "rare_words": args.rare_words,
    }


def main() -> None:
    """Main function, called when the 'correct' command is invoked"""

    args = parser.parse_args()
    # Fill options with information from args
    if args.infile is sys.stdin and sys.stdin.isatty():
        # terminal input is empty, most likely no value was given for infile:
        # Nothing we can do
        print("No input has been given, nothing can be returned")
        sys.exit(1)
    options = from_args(args)

    print(check_errors(**options), file=args.outfile)


if __name__ == "__main__":
    main()
